/* jcguess - jcode guess
 * 1999 (C) Masanao Izumo <mo@goice.co.jp>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>

FILE *in;	/* Input stream to check */

#define IS_ASCII(c)	((0x09 <= (c) && (c) <= 0x0d) || /* isspace(c) */ \
			 (0x20 <= (c) && (c) <= 0x7e)) /* isprint(c) */

typedef struct _JCodeGuessState
{
    char *name;
    void (* handler)(struct _JCodeGuessState *guess, int c);
    int state;
    int len;	/* matched total length */
    int err;	/* error count */

    int first_error;	/* first error position */
} JCodeGuessState;

void jcerror(JCodeGuessState *guess)
{
    guess->err++;
    if(guess->first_error == -1)
	guess->first_error = ftell(in);
}


/*
 * New line detectors
 */
enum
{
    JC_EXPECT_CR,
    JC_EXPECT_LF
};
void guess_cr(JCodeGuessState *guess, int c)
{
    switch(c)
    {
      case '\r':
	guess->len++;
	break;
      case '\n':
	jcerror(guess);
	break;
    }
}
void guess_lf(JCodeGuessState *guess, int c)
{
    switch(c)
    {
      case '\r':
	jcerror(guess);
	break;
      case '\n':
	guess->len++;
	break;
    }
}
void guess_crlf(JCodeGuessState *guess, int c)
{
    switch(guess->state)
    {
      case JC_EXPECT_CR:
	switch(c)
	{
	  case '\r':
	    guess->state = JC_EXPECT_LF;
	    break;
	  case '\n':
	    jcerror(guess);
	    break;
	}
	break;
      case JC_EXPECT_LF:
	switch(c)
	{
	  case '\n':
	    guess->len += 2;
	    break;
	  default:
	    jcerror(guess);
	    break;
	}
	guess->state = JC_EXPECT_CR;
	break;
    }
}
JCodeGuessState cr = {"cr", guess_cr, 0, 0, 0, -1};
JCodeGuessState lf = {"lf", guess_lf, 0, 0, 0, -1};
JCodeGuessState crlf = {"crlf", guess_crlf, JC_EXPECT_CR, 0, 0, -1};


/*
 * ASCII detector
 */

void guess_ascii(JCodeGuessState *guess, int c)
{
    if(IS_ASCII(c))
	guess->len++;
    else
	jcerror(guess);
}
JCodeGuessState ascii = {"ascii", guess_ascii, 0, 0, 0, -1};



/*
 * EUC detector
 */
enum
{
    JC_EUC_INIT,
    JC_EUC_ZENKAKU,
    JC_EUC_HANKAKU,
};
void guess_euc(JCodeGuessState *guess, int c)
{
    switch(guess->state)
    {
      case JC_EUC_INIT:
	if(IS_ASCII(c))
	    guess->len++;
	else if(0xa1 <= c && c <= 0xfe)
	    guess->state = JC_EUC_ZENKAKU;
	else if(c == 0x8e) /* SS2 */
	    guess->state = JC_EUC_HANKAKU;
	else
	{
	    jcerror(guess);
	    guess->state = JC_EUC_INIT;
	}
	break;
	    
      case JC_EUC_ZENKAKU:
	if(0xa1 <= c && c <= 0xfe)
	    guess->len += 2;
	else
	    jcerror(guess);
	guess->state = JC_EUC_INIT;
	break;

      case JC_EUC_HANKAKU:
	if(0xa1 <= c && c <= -0xdf)
	    guess->len += 2;
	else
	    jcerror(guess);
	guess->state = JC_EUC_INIT;
	break;
    }
}
JCodeGuessState euc = {"euc", guess_euc, JC_EUC_INIT, 0, 0, -1};


/*
 * SJIS detector
 */
enum
{
    JC_SJIS_INIT,
    JC_SJIS_ZENKAKU,
};
void guess_sjis(JCodeGuessState *guess, int c)
{
    switch(guess->state)
    {
      case JC_SJIS_INIT:
	if(IS_ASCII(c) || /* ascii */
	   (0xa1 <= c && c <= 0xdf))	/* hankaku */
	    guess->len++;
	else if((0x81 <= c && c <= 0x9f) ||
		(0xe0 <= (c) && (c) <= 0xfc))
	    guess->state = JC_SJIS_ZENKAKU;
	else
	    jcerror(guess);
	break;

      case JC_SJIS_ZENKAKU:
	if((0x40 <= c && c <= 0x7e) ||
	   (0x80 <= c && c <= 0xfc))
	    guess->len += 2;
	else
	    jcerror(guess);
	guess->state = JC_SJIS_INIT;
	break;
    }
}
JCodeGuessState sjis = {"sjis", guess_sjis, JC_SJIS_INIT, 0, 0, -1};

/*
 * ISO-2022-JP detector
 */
enum
{
    JC_JP_ASCII,
    JC_JP_ZENKAKU1,
    JC_JP_ZENKAKU2,
    JC_JP_HANKAKU,
    JC_JP_ESCAPE,
    JC_JP_ESC_D,	/* ESC $ */
    JC_JP_ESC_P,	/* ESC ( */
    JC_JP_SO		/* SO-SI */
};
void guess_iso2022jp(JCodeGuessState *guess, int c)
{
    switch(guess->state)
    {
      case JC_JP_ASCII:
	if(IS_ASCII(c))
	    guess->len++;
	else if(c == 0x1b) /* ESC */
	    guess->state = JC_JP_ESCAPE;
	else
	    jcerror(guess);
	break;

      case JC_JP_ESCAPE:
	switch(c)
	{
	  case '$':
	    guess->state = JC_JP_ESC_D;
	    break;
	  case '(':
	    guess->state = JC_JP_ESC_P;
	    break;
	  default:
	    jcerror(guess);
	    guess->state = JC_JP_ASCII;
	    break;
	}
	break;

      case JC_JP_ESC_D:
	switch(c)
	{
	  case '@':
	  case 'B':
	    guess->len += 3;
	    guess->state = JC_JP_ZENKAKU1;
	    break;
	  default:
	    jcerror(guess);
	    guess->state = JC_JP_ASCII;
	    break;
	}
	break;

      case JC_JP_ESC_P:
	switch(c)
	{
	  case 'J':
	  case 'B':
	    guess->len += 3;
	    guess->state = JC_JP_ASCII;
	    break;
	  case 'I':
	    guess->len += 3;
	    guess->state = JC_JP_HANKAKU;
	    break;
	  default:
	    jcerror(guess);
	    guess->state = JC_JP_ASCII;
	    break;
	}
	break;

      case JC_JP_ZENKAKU1:
	if(0x21 <= c && c <= 0x7e)
	    guess->state = JC_JP_ZENKAKU2;
	else if(c == 0x1b) /* ESC */
	    guess->state = JC_JP_ESCAPE;
	else if(c == 0x0e)
	{
	    guess->len++;
	    guess->state = JC_JP_SO;
	}
	else
	{
	    jcerror(guess);
	    guess->state = JC_JP_ASCII;
	}
	break;

      case JC_JP_ZENKAKU2:
	if(0x21 <= c && c <= 0x7e)
	{
	    guess->len += 2;
	    guess->state = JC_JP_ZENKAKU1;
	}
	else
	{
	    jcerror(guess);
	    guess->state = JC_JP_ASCII;
	}
	break;

      case JC_JP_SO:
	if(0x21 <= c && c <= 0x5f)
	    guess->len++;
	else if(c == 0x0f)
	{
	    guess->len++;
	    guess->state = JC_JP_ZENKAKU1;
	}
	else
	{
	    jcerror(guess);
	    guess->state = JC_JP_ASCII;
	}
	break;
    }
}
JCodeGuessState iso2022jp = {"2022jp", guess_iso2022jp, JC_JP_ASCII, 0, 0, -1};

void header(char *filename)
{
    printf("[%s]:\n", filename);
    printf("<CODE>\t<MATCH>\t<ERROR>\t<RATE>\t<ERRPOS>\n");
}

void doit(void)
{
    JCodeGuessState *states[] = {
	&ascii, &euc, &sjis, &iso2022jp,
	&cr, &lf, &crlf};
#define NUM_GUESS (sizeof(states) / sizeof(states[0]))
    char save[sizeof(JCodeGuessState) * NUM_GUESS];
    int i;
    int c;

    /* save initial states */
    for(i = 0; i < NUM_GUESS; i++)
	memcpy(save + i * sizeof(JCodeGuessState),
	       states[i],
	       sizeof(JCodeGuessState));

    /* doit */
    while((c = getc(in)) != EOF)
    {
	for(i = 0; i < NUM_GUESS; i++)
	    states[i]->handler(states[i], c);
    }

    /* report */
    for(i = 0; i < NUM_GUESS; i++)
    {
	printf("%s", states[i]->name);
	printf("\t%d", states[i]->len);
	printf("\t%d", states[i]->err);
	if(states[i]->len + states[i]->err > 0)
	    printf("\t%d%%",
		   states[i]->len * 100 / (states[i]->len + states[i]->err));
	else
	    printf("\t?%%");

	printf("\t");
	if(states[i]->err)
	    printf("%d", states[i]->first_error);
	else
	    printf("-");

	printf("\n");
    }

    /* restore initial states */
    for(i = 0; i < NUM_GUESS; i++)
	memcpy(states[i],
	       save + i * sizeof(JCodeGuessState),
	       sizeof(JCodeGuessState));
}

int main(int argc, char **argv)
{
    if(argc <= 1)
    {
	in = stdin;
	header("-");
	doit();
    }
    else
    {
	int i;
	for(i = 1; i < argc; i++)
	{
	    if((in = fopen(argv[i], "r")) == NULL)
		perror(argv[i]);
	    else
	    {
		header(argv[i]);
		doit();
		fclose(in);
	    }
	}
    }
    return 0;
}
