#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <ctype.h>
#include <limits.h>		/* for UCHAR_MAX */

int expand_string(char* string);
char* readfile(const char* name, size_t* size);
char* readtty(size_t* size);
int euc_mode = 0;
void  print_replace_string(const char* text, size_t text_size,
			   const char* from, int from_len,
			   const char* to, int to_len);

void usage()
{
    fprintf(stderr, "usage: replace-string from to [-e] [ filename ... ]\n");
    fprintf(stderr, "       -e option for euc-code\n");
}

int main(int argc, char** argv)
{
    char* from;
    char* to;
    char* text;
    size_t text_size;
    int i, from_len, to_len;

    if(argc < 3)
    {
	usage();
	return 1;
    }

    from = argv[1];
    to   = argv[2];

    if((from_len = expand_string(from)) == 0)
    {
	fprintf(stderr, "from string is empty.\n");
	return 1;
    }

    to_len = expand_string(to);

    i = 3;
    if(i < argc && !strcmp(argv[i], "-e"))
    {
	euc_mode = 1;
	i++;
    }

    if(argc == i)
    {
	text = readtty(&text_size);
	print_replace_string(text, text_size,
			     from, from_len, to, to_len);
    }
    else
    {
	for(; i < argc; i++)
	{
	    char* fname;

	    fname = argv[i];
	    if(!strcmp(fname, "-"))
	    {
		text = readtty(&text_size);
		print_replace_string(text, text_size,
				     from, from_len, to, to_len);
		free(text);
	    }
	    else
	    {
		if((text = readfile(fname, &text_size)) == NULL)
		    continue;
		print_replace_string(text, text_size,
				     from, from_len, to, to_len);
		munmap(text, text_size);
	    }
	}
    }
    return 0;
}

int expand_string(char* from)
{
    char* to;
    char* to_start;

    to_start = to = from;
    while(*from)
    {
	if(*from == '\\')
	{
	    from++;
	    if(*from == '\0')
		break;
            switch(*from)
	    {
#define EXPAND(a, b) case a: *to++ = b; from++; break;
                EXPAND('n', '\n');
                EXPAND('t', '\t');
                EXPAND('v', '\v');
                EXPAND('b', '\b');
                EXPAND('r', '\r');
                EXPAND('f', '\f');
                EXPAND('a', '\a');
                EXPAND('\\', '\\');
#undef EXPAND
	      case '0':
	      case '1':
	      case '2':
	      case '3':
		if(isdigit(from[1]) && isdigit(from[2]))
		{
		    *to++ = (from[0] - '0') * 64
			  + (from[1] - '0') * 8
			  + (from[2] - '0');
		    from += 3;
		    break;
		}
		else if(from[0] == '0')
		{
		    *to++ = '\0';
		    from++;
		    break;
		}

		/* THROUGH */
	      default:
		*to++ = *from++;
		break;
	    }
	}
	else
	    *to++ = *from++;
    }
    *to = '\0';
    return to - to_start;
}

char* readfile(const char* name, size_t* size)
{
    int fd;
    struct stat buf;
    char* buff;

    if((fd = open(name, O_RDONLY)) < 0)
    {
	perror(name);
	return NULL;
    }

    if(fstat(fd, &buf) < 0)
    {
	perror(name);
	close(fd);
	return NULL;
    }
    *size = buf.st_size;

    if((buff = mmap(0, *size, PROT_READ, MAP_SHARED, fd, 0)) == (caddr_t) - 1)
    {
	perror(name);
	close(fd);
	return NULL;
    }
    close(fd);

    return buff;
}



/* readtty */
#define UNIT_SIZE 1048576	/* 1M */
#define READ_UNIT (1024*64)	/* 64k */

struct list
{
    char* buffer;
    int   n;
    struct list* next;
};

static struct list* first = NULL;
static struct list* last = NULL;

static struct list* newlist(void)
{
    struct list* node;

    if((node = (struct list *)malloc(sizeof(struct list))) == NULL ||
       (node->buffer = (char *)malloc(UNIT_SIZE)) == NULL)
    {
	fprintf(stderr, "Can't allocate memory\n");
	exit(1);
    }

    node->n = 0;
    node->next = NULL;

    return node;
}

static void next_string(char* s, int n)
{
    struct list* p;

    if(first == NULL)
	last = first = newlist();
    p = last;
    if(p->n + n < UNIT_SIZE)
    {
	memcpy(p->buffer + p->n, s, n);
	p->n += n;
    }
    else
    {
	int n0;

	n0 = UNIT_SIZE - p->n;
	memcpy(p->buffer + p->n, s, n0);
	p->n = UNIT_SIZE;
	last = p->next = newlist();
	p = last;
	memcpy(p->buffer, s + n0, n - n0);
	p->n = n - n0;
    }
}

static char* concat_string(char* s, size_t* nread)
{
    struct list* p;
    struct list* cur;
    char* start = s;

    cur = first;
    *nread = 0;
    while(cur != NULL)
    {
	p = cur;
	cur = cur->next;
	memcpy(s, p->buffer, p->n);
	s += p->n;
	*nread += p->n;
	free(p->buffer);
	free(p);
    }

    last = first = NULL;
    start[*nread] = '\0';
    return start;
}

char* readtty(size_t* nread)
{
    char buff[READ_UNIT];
    int fd = 0;
    long n;
    char* s;
    size_t t;

    t = 0;
    while((n = read(fd, buff, READ_UNIT)) > 0)
    {
	t += n;
	next_string(buff, n);
    }

    if(n < 0)
    {
	perror("read");
	exit(1);
    }

    if((s = (char *)malloc(t + 1)) == NULL)
    {
	fprintf(stderr, "Can't allocate memory\n");
	exit(1);
    }

    return concat_string(s, nread);
}
/* end of readtty */

void make_bm_skip(const unsigned char* pattern, long pattern_len, long* skip);
long BM_search(unsigned char* text,    long text_len,
	       unsigned char* pattern, long pattern_len,
	       long* skip);

void  print_replace_string(const char* text, size_t text_size,
			   const char* from, int from_len,
			   const char* to, int to_len)
{
    long skip[UCHAR_MAX + 1];
    long offset;

    make_bm_skip((const unsigned char *)from, from_len, skip);

    while(1)
    {
	offset = BM_search((unsigned char *)text, text_size,
			   (unsigned char *)from, from_len,
			   skip);
	if(offset == -1)
	{
	    write(1, text, text_size);
	    break;
	}
	else
	{
	    write(1, text, offset);
	    write(1, to, to_len);
	    text += offset + from_len;
	    text_size -= offset + from_len;
	}
    }
}

void make_bm_skip(const unsigned char* pattern, long len, long* skip)
{
    long i;

    /* make the skip table */
    for(i = 0; i <= UCHAR_MAX; i++)
	skip[i] = len;
    for(i = 0; i < len - 1; i++)
	skip[pattern[i]] = len - 1 - i;
}

int boeucwcp(unsigned const char* body, long pos)
{
    while(pos > 0 && (body[pos] & 0x80))
    {
	if((body[pos - 1] & 0x80) == 0)
	    return 1;
	pos -= 2;
    }
    return pos == 0;
}

long BM_search(unsigned char* text,    long tlen,
	       unsigned char* pattern, long len,
	       long* skip)
{
    long i, j, k;
    unsigned char c, tail;

    if(len == 0)
	return -1;

    if(len == tlen)
    {
	if(strcmp((const char *)text, (const char *)pattern) == 0)
	    return 0;
	return -1;
    }
    tail = pattern[len - 1];

    if(len == 1)
    {
	for(i = 0; i < tlen; i++)
	    if(text[i] == tail)
		return i;
    }
    else
    {
	/* matching procedure */
	i = len - 1;
	while(i < tlen)
	{
	    c = text[i];
	    if(c == tail)
	    {
		j = len - 1;	/* index of pattern */
		k = i;		/* index of text */
		while(pattern[--j] == text[--k])
		    if(j == 0)
		    {
			if((text[k] & 0x80) && !boeucwcp(text, k))
			    break;
			return k; /* found */
		    }
	    }
	    i += skip[c];
	}
    }
    return -1;
}
